from logging import getLogger
from random import randrange
from rainbowrooms.generator import generate
import libtcodpy as tcod

log = getLogger("rainbowrooms.world")

def neighborhood(x,y):
    result = set(
        (x+dx, y+dy)
        for dx in (-1, 0, 1)
        for dy in (-1, 0, 1)
    )
    
    result.discard((x,y))
    return result

class Tile(object):
    symbol = None
    walkable = False
    passable = False
    transparent = True
    color = tcod.darkest_grey
    
    @classmethod
    def bump(self):
        pass

class Space(Tile):
    colors = tcod.color_gen_map([tcod.darkest_grey, tcod.dark_grey], [0, 10])
    walkable = True
    
    def __init__(self, levels, has_player=False):
        self.passable = levels
        self.travelled = has_player
    
    def bump(self):
        if self.travelled < len(self.colors) - 1:
            self.travelled += 1
    
    @property
    def color(self):
        return self.colors[self.travelled]

class Wall(Tile):
    switched = [True] * 5
    colors = [
        (tcod.darkest_flame,   tcod.flame),
        (tcod.darkest_yellow,  tcod.yellow),
        (tcod.darkest_sea,     tcod.sea),
        (tcod.darkest_azure,   tcod.azure),
        (tcod.darkest_fuchsia, tcod.fuchsia),
    ]
    
    def __init__(self, levels, index):
        self.index = index
        self.z = levels
    
    @property
    def color(self):
        lit = self.switched[self.index]
        return self.colors[self.index][lit]
    
    @property
    def passable(self):
        if self.switched[self.index]:
            return False
        else:
            return self.z
    
    @property
    def transparent(self):
        return not self.switched[self.index]
    walkable = transparent

class Switch(Wall):
    symbol = tcod.CHAR_BULLET
    walkable = False
    passable = False
    transparent = True
    colors = [
        (tcod.darker_flame,   tcod.flame),
        (tcod.darker_yellow,  tcod.yellow),
        (tcod.darker_sea,     tcod.sea),
        (tcod.darker_azure,   tcod.azure),
        (tcod.darker_fuchsia, tcod.fuchsia),
    ]
    
    def bump(self):
        self.switched[self.index] = not self.switched[self.index]
        self.switched[(self.index + 2) % 5] = True
        self.switched[(self.index + 3) % 5] = True

class Goal(Tile):
    symbol = tcod.CHAR_HEART
    color = tcod.red
    
    def __init__(self, levels):
        self.passable = levels
    
    def bump(self):
        self.symbol = tcod.CHAR_LIGHT
        self.color = tcod.yellow
        return True

class Rock(Tile):
    passable = False
    transparent = False
    color = tcod.sepia

class Level(object):
    def __init__(self, size, z, previous):
        self.spaces = {}
        self.size = size
        self.start = (2, 2)
        self.load(z, previous)
    
    def load(self, z, previous):
        log.debug("Loading level %s", z)
        width, height, depth = self.size
        bottom = (z == depth)
        colors, cells, walls, self.state = generate(width, height, bottom, previous and previous.state)
        
        self.z = (z,)
        self.down = (z, z+1)
        self.above = previous
        self.static = {}
        self.static["$"] = Goal(self.z)
        for index, char in enumerate(colors):
            self.static[char.upper()] = Wall(self.z, index)
            self.static[char.lower()] = Switch(self.z, index)
        
        for cell in cells:
            x, y = cell
            self.room(x<<1, y<<1, cells[cell])
        
        handlers = [self.hall_vert, self.hall_horiz]
        for wall in walls:
            x, y, vertical = wall
            x <<= 1
            y <<= 1
            if vertical:
                x += 1
            else:
                y += 1
            
            char = (walls[wall] or " ").upper()
            handlers[vertical](x, y, char)
    
    def room(self, x, y, char):
        x = (x<<1) + 2
        y = (y<<1) + 2
        
        tile = None
        if char == "@":
            self.start = (x, y)
            tile = Space(self.z, 1)
        elif char == "+":
            tile = Space(self.z)
        elif char in self.static:
            tile = self.static.get(char)
        else:
            # Empty room.
            return
        
        self.spaces[x,y] = tile
        for dx,dy in neighborhood(x,y):
            self.spaces[dx,dy] = Space(self.z)
    
    def hall_horiz(self, x, y, char):
        x = (x<<1) + 2
        y = (y<<1) + 1
        self.hallway(x, y, char, [0], [0, 1, 2])
    
    def hall_vert(self, x, y, char):
        x = (x<<1) + 1
        y = (y<<1) + 2
        self.hallway(x, y, char, [0, 1, 2], [0])
    
    def hallway(self, x, y, char, dxs, dys):
        tile = None
        if char == "+":
            levels = self.z
        elif char == ">":
            levels = self.down
        elif char == "<":
            z = self.z[0]
            up = (z-1, z)
            for dx in dxs:
                for dy in dys:
                    self.spaces[x+dx,y+dy] = tile = self.above.spaces[x+dx,y+dy]
                    assert tile.passable == up
            return
        else:
            tile = self.static.get(char)
            if not tile:
                return
        
        for dx in dxs:
            for dy in dys:
                self.spaces[x+dx,y+dy] = tile or Space(levels)
    
    def ignore(self, x, y, char):
        pass

class World(object):
    def __init__(self):
        self.maze = 0
        self.size = [4,3,0]
        self.reset()
    
    def reset(self):
        self.maze += 1
        if self.maze > 42:
            # Increase only depth once we fill the screen.
            self.size[2] += 1
        elif not self.maze % 6:
            # Increase depth every six mazes.
            self.size[2] += 1
        elif not self.maze % 2:
            # Increase height two out of every six mazes.
            self.size[1] += 1
        else:
            # Increase width three out of every six mazes.
            self.size[0] += 1
        log.info("Maze #%d: %s", self.maze, self.size)
        
        level = Level(self.size, 0, None)
        x, y = level.start
        z = (0,)
        self.levels = [level]
        self.player = (x, y, z)
    
    def __getitem__(self, loc):
        x, y, levels = loc
        for z in levels:
            space = self.level(z).spaces.get((x,y))
            if space:
                return space
        return Rock
    
    def level(self, z):
        maxlev = len(self.levels)
        assert 0 <= z <= maxlev
        if z == maxlev:
            level = Level(self.size, z, self.levels[-1])
            self.levels.append(level)
        else:
            level = self.levels[z]
        return level
